//
// QSINIT
// internal loader part defines (not available for other code)
//
#ifndef QSINIT_LOCAL_HEADER
#define QSINIT_LOCAL_HEADER

#include "qsmodint.h"
#include "qsbase.h"
#include "qsinit.h"
#include "qspdata.h"
#include "qsstor.h"
#include "qserr.h"
#include "qsint.h"
#include "memmgr.h"       // QSMEMOWNER_* constants

#define CALLER(x) (((u32t*)(&x))[-1])

typedef  _std (*pf_mtpexitcb)(mt_thrdata *td, int stage);

void  _std usleep      (u32t usec);
void  _std mt_yield    (void);
u64t  _std hlp_tscread (void);
int   _std sys_intstate(int on);

u16t  _std key_read_nw (u32t *status);

int   _std tolower     (int cc);
u32t  _std strlen      (const char *str);
u32t  _std wcslen      (const u16t *str);
u16t       ff_convert  (u16t chr, unsigned int dir);
u32t       env_length  (const char *env);

void       mem_procexit(process_context *pq);
u32t  _std hlp_fopen2  (const char *name, int allow_pxe);
// serial port init, has no lock inside!
u32t  _std serial_init (u16t port);
void       init_host   (void);

/// secondary function table, from "start" module
extern mod_addfunc *mod_secondary; 
extern pf_mtpexitcb    mt_pexitcb;
extern volatile 
       mt_proc_cb    mt_exechooks;
extern vol_data*           extvol;
/// disk buffer address (see also DiskBufRM_Seg in BIOS host)
extern u32t             DiskBufPM;
/// boot partition disk i/o is available
extern u8t           bootio_avail,
                         pxemicro,
                         streamio,
                         safeMode,
                      mfsd_noterm,
                           vmmode;

extern u16t           ComPortAddr;
extern u32t              BaudRate,
                        bootmount;

extern u32t         cvio_ttylines;
extern void             *init_env;

extern char             *ldi_name;

/** host mode 4k log buffer.
    Used for real mode logging in BIOS host, and for 64-bit code logging in
    UEFI, contents commited to the system log after return to 32-bit home
    world ;) */
extern u8t*              logrmbuf;

extern u16t                cp_num;

/* is virtual memory addressing active?
   Page mode can be actually - old way PAE with FLAT mapping or PAE/32-bit
   with modules mapped at the top of the memory.
   This macro signals about the second */
#define VMA()    (mt_exechooks.mtcb_pgmode & QPGM_VMA)

#ifndef EFI_BUILD
/// check/reset BIOS Ctrl-Break flag in 040:0071h
int   _std check_cbreak(void);
/// boot flags
extern u8t           dd_bootflags,
                      dd_bootdisk;
extern u32t        transition_ptr;  ///< ecx value at loader start
extern u32t              qs16base;  ///< base addr of 16-bit part
extern void          *minifsd_ptr;  ///< buffer with mini-fsd or 0
extern u16t             logbufseg,  ///< real mode 4k log buffer
                    DiskBufRM_Seg;  ///< disk buffer real mode segment
extern u64t            page0_fptr;

/** push fixups for switching addresses to page mode.
    @param  addr   pointer to the memory block address (block should be
                   allocated via hlp_memalloc())
    @param  nsub   number of optional subpointers, each must be within this
                   block */
void __cdecl     push_fixup(void *addr, u32t nsub, ...);
void             vm_panic  (void);

u32t             vm_reserve(u32t physbase, u32t size, u32t flags);
u32t             vm_realloc(u32t addr, u32t newsize);
void             vm_release(u32t addr);
u32t             vm_vtophys(u32t vmaddr);
u32t             vm_avail  (u32t *maxblock, u32t *total);

void             init_pages(void);
#pragma aux init_pages aborts;

// this should be used for all 16-bit code rmcall()-s in this module ...
u32t __cdecl     rmcall_loc(u32t rmfunc, u32t dwcopy, ...);
#define rmcall(func,...) rmcall_loc((u32t)&(func),__VA_ARGS__)

#define VMBASE   ((u32t)-INIT_SYS_AREA)  ///< base address of VM area

#else // EFI_BUILD
#define push_fixup(addr,...)
extern char             *ini_path;  ///< custom INI file path
#endif

void             call_start(void);
#pragma aux call_start aborts;

void  _std       sys_setcr3(u32t syscr3, u32t syscr4);
u32t             getcr4    (void);
#ifdef __WATCOMC__
#pragma aux getcr4 = "mov eax,cr4" modify exact [eax];
#endif

/// get current thread data
mt_thrdata* _std mt_curthread(void);

/* - - - - - - - - - - - - - - - - - */
typedef struct {
   u32t                  sign;
   u32t                 crc32;
   u32t                length;
   u32t                  type;
} transition_data;

#define TRND_ENV                1         ///< block with env. data
#define TRND_CMDHIST            2         ///< block with shell history list
#define TRND_CODEPAGE           3         ///< codepage selected
#define TRND_FLAGS              4         ///< various flags
#define TRND_CFGFILE            5         ///< INI file contents
#define TRND_LAST      0x80000000         ///< last block flag
#define TRND_SIGN      0x444E5254

#define next_trblock(td) ((transition_data*)((u8t*)(td+1) + Round16(td->length)))

#endif // QSINIT_LOCAL_HEADER
