//
// QSINIT "start" module
// special heap for arena management code
//
#ifndef QSINIT_PTHEAP
#define QSINIT_PTHEAP

#include "qstypes.h"

#ifdef __cplusplus
extern "C" {
#endif
typedef void *ptheap;

/** initialize heap.
    Heap internal data uses global memory (with optional QSMA_READONLY),
    heap blocks use local or global - depends on QSMA_LOCAL).

    @param  blocksize   size of the heap block
    @param  limit       number of heap blocks
    @param  memflags    only QSMA_LOCAL and QSMA_READONLY are accepted (r/o
                        can be used for PID 1 heaps)
    @param  [out] res   heap if success
    @return error code or 0 */
qserr   pth_init (u32t blocksize, u32t limit, u32t memflags, ptheap *res);

/** free heap.
    Caller must free the heap since there is no auto-free at process exit
    (except the local owned memory).
    Function can be called from another process context only with freemem=0.
    @param  hh          heap
    @param  freemem     free allocated memory (1/0)
    @return error code or 0 */
qserr   pth_fini (ptheap hh, int freemem);

/** allocate a block in the heap.
    @param  hh          heap
    @return block index or 0 if error */
u32t    pth_alloc(ptheap hh);

/** get pointer to the index.
    returned pointer is constant while block allocated, caller process
    gets local pointer if QSMA_LOCAL is used, any other - mapped in the
    system arena (and this is bad).

    @param  hh          heap
    @param  index       index to get (should be allocated)
    @return pointer or 0 */
void   *pth_get  (ptheap hh, u32t index);

/// free the heap block
qserr   pth_free (ptheap hh, u32t index);

#ifdef __cplusplus
}
#endif
#endif // QSINIT_PTHEAP
