//
// QSINIT EFI loader
// log & minor runtime
//
#include "qsloader.h"
#include "efilib.h"
#include "efistdarg.h"

// use internal EFI Lib call to make life more friendly
VOID _PoolCatPrint (IN CHAR16 *fmt, IN va_list args, IN OUT POOL_PRINT *spc,
   IN INTN (*Output)(VOID *context, CHAR16 *str));

INTN _SPrint (IN VOID *Context, IN CHAR16 *Buffer);

#define LOGIF_LEVELMASK 0x0003 ///< mask for priority value
#define LOGIF_SECOND    0x0004 ///< 1 second bit for dostime
#define LOGIF_DELAY     0x0010 ///< entry was delayed
#define LOGIF_REALMODE  0x0020 ///< entry was created in real mode
#define LOGIF_UEFICODE  0x0040 ///< entry was created in 64-bit code

int log_it(int level, CHAR16 *fmt, ...) {
   if (!Header32) return 0; else {
      va_list     args;
      POOL_PRINT   spc;
      char         *cp = (char*)(u64t)Header32->logrmbuf;
      u32t         pos = Header32->logrmpos;
      CHAR16    buffer[128], *cb = buffer;
      
      if (!cp || pos>=LOGBUF_SIZE-5-4) return 0;
      // print to buffer
      va_start (args, fmt);
      spc.str    = buffer;
      spc.maxlen = 128 - 1;
      spc.len    = 0;
      _PoolCatPrint (fmt, args, &spc, _SPrint);
      // level 0..3 and LOGIF_DELAY+LOGIF_UEFICODE flags
      level = level&LOGIF_LEVELMASK|LOGIF_DELAY+LOGIF_UEFICODE;
      /* if there is no EOL at the end of last entry - appends this line to it,
         else starts a new entry */
      if (!pos || cp[pos-1]=='\n') {
         u64t time = tm_getdate();
         if (pos) pos++;
         if (time&1) level|=LOGIF_SECOND;
         cp[pos++] = (u8t)level;
         /* save dos time.
            this is mandatory action, else log_flush() on exiting from 64-bit
            part will ask for current time */
         *(u32t*)(cp+pos) = (u32t)(time>>1); pos+=4;
      }
      while (pos<LOGBUF_SIZE-1)
         if (*cb=='\r') cb++; else
            if ((cp[pos] = (char)*cb++)==0) break; else pos++;
      // logrmpos must point to the term NULL in the last string!
      Header32->logrmpos = (u16t)pos;
   }
   return 1;
}

/// stricmp for A..Z chars
int RtStrICmp(const CHAR16 *s1, const CHAR16 *s2) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (1) {
      CHAR16 ch1 = *s1++,
             ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
}

/// strnicmp for A..Z chars
int RtStrNICmp(const CHAR16 *s1, const CHAR16 *s2, UINTN len) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (len--) {
      CHAR16 ch1 = *s1++,
             ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
   return 0;
}

int strnicmp(const char *s1, const char *s2, u32t len) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (len--) {
      char ch1 = *s1++,
           ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
   return 0;
}

CHAR16* wcscat(CHAR16 *dst, const CHAR16 *src) {
   wcscpy(dst+wcslen(dst), src);
   return dst;
}

CHAR16 *strto16(const char *src, UINTN maxlen) {
   CHAR16 *res;
   UINTN   len = strlen((char*)src), ii;
   if (maxlen && len>maxlen) len = maxlen;

   res = (CHAR16*)AllocatePool(len+1);
   if (!res) return 0;

   for (ii=0; ii<len; ii++) res[ii] = src[ii];
   res[len] = 0;
   return res;
}

void DumpHandleProtocols(const CHAR16 *info, EFI_HANDLE *eh) {
   EFI_STATUS    rc;
   EFI_GUID    **ga;
   UINTN      count;

   rc = BS->ProtocolsPerHandle(eh, &ga, &count);
   if (!EFI_ERROR(rc)) {
      EFI_DEVICE_PATH  *dp;
      UINTN           pidx;

      log_it(3, L"%s -> handle %x: %d protocols\n", info, eh, count);
      dp = DevicePathFromHandle(eh);
      if (dp) {
         CHAR16 *dpstr = DevicePathToStr(dp);
         log_it(3, L"device path: %s\n", dpstr);
         BS->FreePool(dpstr);
      }
      for (pidx=0; pidx<count; pidx++) log_it(3, L"%g\n", ga[pidx]);

      BS->FreePool(ga);
   } else {
      log_it(3, L"%s -> handle %x: %r\n", info, eh, rc);
   }
}
